"""
spall.exceptions
================

Exceptions for use within the module.

All exceptions made public for if they need to be reraised or excepted.
"""
# noinspection PyUnresolvedReferences
from subprocess import CalledProcessError  # noqa pylint: disable=unused-import


class CommandNotFoundError(OSError):
    """Raise when subprocess called is not on system.

    :param cmd: Name of command that was not found.
    """

    def __init__(self, cmd: str) -> None:
        super().__init__(f"{cmd}: command not found...")



"""
spall._subprocess
=================
"""
from __future__ import annotations

import functools as _functools
import os as _os
import subprocess as _sp
import sys as _sys
import typing as _t

from . import exceptions as _exceptions


class _Stream:
    def __init__(self) -> None:
        self._list: _t.List[str] = []

    def consume(self) -> _t.List[str]:
        """Consume and return accrued output.

        :return: List of captured output.
        """
        captured, self._list = self._list, []
        return captured

    def append(self, item: str) -> None:
        """Append to ``_list``.

        :param item: Output as str.
        """
        self._list.append(item)


class Subprocess:
    """Create a new ``Subprocess`` object.

    ``cmd`` is a mandatory argument used to construct the subprocess
    executable.

    Default ``file``, ``suppress``, and ``capture`` values can be set
    when instantiating the object to be overridden later when using
    ``call``, or simply set through the ``call`` method alone.

    :param cmd: Subprocess executable.
    :param positionals: List of positional arguments to set as
        attributes if not None.
    :key file: File path to write stdout and stderr to if not None.
    :key file_stdout: File path to write stdout to if not None.
    :key file_stderr: File path to write stderr to if not None.
    :key capture: Collect stdout and stderr array.
    :key capture_stdout: Collect stdout array.
    :key capture_stderr: Collect stderr array.
    :key pipe: Pipe stderr to stdout.
    :key suppress: Suppress errors and continue running.
    """

    def __init__(
        self,
        cmd: str,
        positionals: _t.Iterable[str] | None = None,
        **kwargs: bool | str | _os.PathLike,
    ) -> None:
        self._cmd = cmd
        self._kwargs = kwargs
        self._stdout = _Stream()
        self._stderr = _Stream()
        if positionals is not None:
            self._set_positionals(positionals)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} ({self._cmd})>"

    def __setattr__(self, key: str, value: _t.Any) -> None:
        # pylint: disable=useless-super-delegation
        # not useless; silences `mypy` with `_t.Any` and dynamic attrs
        super().__setattr__(key, value)

    def __getattribute__(self, key: str) -> _t.Any:
        # pylint: disable=useless-super-delegation
        # not useless; silences `mypy` with `_t.Any` and dynamic attrs
        return super().__getattribute__(key)

    def _set_positionals(self, positionals: _t.Iterable[str]) -> None:
        for positional in positionals:
            setattr(
                self,
                positional.replace("-", "_"),
                _functools.partial(self.call, positional),
            )

    def _handle_stream(
        self, popen: _sp.Popen, std: str, **kwargs: bool | str | _os.PathLike
    ) -> None:
        std_pipe = getattr(popen, std)
        if std_pipe is not None:
            for line in iter(std_pipe.readline, b""):
                if kwargs.get("pipe", self._kwargs.get("pipe", False)):
                    std = "stdout"

                line = line.decode("utf-8", "ignore")
                file = kwargs.get(
                    f"{std}_file",
                    self._kwargs.get(
                        f"{std}_file",
                        kwargs.get("file", self._kwargs.get("file")),
                    ),
                )
                capture = kwargs.get(
                    f"{std}_capture",
                    self._kwargs.get(
                        f"{std}_capture",
                        kwargs.get(
                            "capture", self._kwargs.get("capture", False)
                        ),
                    ),
                )
                if file is not None:
                    with open(file, "a+", encoding="utf-8") as fout:
                        fout.write(line)

                elif capture:
                    getattr(self, f"_{std}").append(line.strip())

                else:
                    sys_std = getattr(_sys, std)
                    if sys_std is not None:
                        sys_std.write(line)

    def _open_process(
        self, *args: str, **kwargs: bool | str | _os.PathLike
    ) -> int:
        cmd = [self._cmd, *args]
        with _sp.Popen(cmd, stdout=_sp.PIPE, stderr=_sp.PIPE) as popen:
            for std in ("stdout", "stderr"):
                self._handle_stream(popen, std, **kwargs)

            return popen.wait()

    def call(self, *args: _t.Any, **kwargs: bool | str | _os.PathLike) -> int:
        """Call command.

        Open process with ``subprocess.Popen``.

        Pipe stream depending on the keyword arguments provided to
        instance constructor or overridden through this method.

        If a file path is provided it will take precedence over
        ``capture``.

        :param args: Positional str arguments.
        :key file: File path to write stdout and stderr to if not None.
        :key file_stdout: File path to write stdout to if not None.
        :key file_stderr: File path to write stderr to if not None.
        :key capture: Collect stdout and stderr array.
        :key capture_stdout: Collect stdout array.
        :key capture_stderr: Collect stderr array.
        :key pipe: Pipe stderr to stdout.
        :key suppress: Suppress errors and continue running.
        :raises CommandNotFoundError: If instantiated executable is not
            in path.
        :raises CalledProcessError: If error occurs in subprocess.
        :return: Exit status.
        """
        args = tuple(str(i) for i in args)
        try:
            returncode = self._open_process(*args, **kwargs)
        except FileNotFoundError as err:
            raise _exceptions.CommandNotFoundError(self._cmd) from err

        if returncode and not kwargs.get(
            "suppress", self._kwargs.get("suppress", False)
        ):
            raise _exceptions.CalledProcessError(
                returncode, " ".join([self._cmd, *args])
            )

        return returncode

    def stdout(self) -> _t.List[str]:
        """Consume accrued stdout by returning the lines of output.

        Assign new container to ``_stdout``.

        :return: List of captured stdout.
        """
        return self._stdout.consume()

    def stderr(self) -> _t.List[str]:
        """Consume accrued stderr by returning the lines of output.

        Assign new container to ``_stderr``.

        :return: List of captured stderr.
        """
        return self._stderr.consume()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['spall']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'spall',
    'version': '0.6.0',
    'description': 'Object-oriented commandline',
    'long_description': 'spall\n=====\n.. image:: https://img.shields.io/badge/License-MIT-yellow.svg\n    :target: https://opensource.org/licenses/MIT\n    :alt: License\n.. image:: https://img.shields.io/pypi/v/spall\n    :target: https://pypi.org/project/spall/\n    :alt: PyPI\n.. image:: https://github.com/jshwi/spall/actions/workflows/ci.yml/badge.svg\n    :target: https://github.com/jshwi/spall/actions/workflows/ci.yml\n    :alt: CI\n.. image:: https://results.pre-commit.ci/badge/github/jshwi/spall/master.svg\n   :target: https://results.pre-commit.ci/latest/github/jshwi/spall/master\n   :alt: pre-commit.ci status\n.. image:: https://github.com/jshwi/spall/actions/workflows/codeql-analysis.yml/badge.svg\n    :target: https://github.com/jshwi/spall/actions/workflows/codeql-analysis.yml\n    :alt: CodeQL\n.. image:: https://codecov.io/gh/jshwi/spall/branch/master/graph/badge.svg\n    :target: https://codecov.io/gh/jshwi/spall\n    :alt: codecov.io\n.. image:: https://readthedocs.org/projects/spall/badge/?version=latest\n    :target: https://spall.readthedocs.io/en/latest/?badge=latest\n    :alt: readthedocs.org\n.. image:: https://img.shields.io/badge/python-3.8-blue.svg\n    :target: https://www.python.org/downloads/release/python-380\n    :alt: python3.8\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://github.com/psf/black\n    :alt: Black\n.. image:: https://img.shields.io/badge/linting-pylint-yellowgreen\n    :target: https://github.com/PyCQA/pylint\n    :alt: pylint\n\nObject-oriented commandline\n---------------------------\n\n\nInstall\n-------\n\n.. code-block:: console\n\n    $ pip install spall\n\nDevelopment\n-----------\n\n.. code-block:: console\n\n    $ pip install spall\n\nUsage\n-----\n\nImport ``Subprocess`` from ``spall``\n\n.. code-block:: python\n\n    >>> from spall import Subprocess\n\nInstantiate individual executables\n\n.. code-block:: python\n\n    >>> cat = Subprocess("cat")\n    >>> echo = Subprocess("echo")\n    >>> fails = Subprocess("false")\n\n\nDefault is to return returncode and print stdout and stderr to console\n\n.. code-block:: python\n\n    >>> returncode = echo.call("Hello, world")\n    Hello, world\n    >>> returncode\n    0\n\nCapture stdout with the ``capture`` keyword argument\n\n.. code-block:: python\n\n    >>> echo.call("Hello, world", capture=True)\n    0\n\nStdout is consumed by calling ``stdout()`` which returns a list\n\n.. code-block:: python\n\n    >>> echo.stdout()\n    [\'Hello, world\']\n    >>> echo.stdout()\n    []\n\nStdout is accrued until ``stdout()`` is called\n\n.. code-block:: python\n\n    >>> echo.call("Hello, world", capture=True)\n    0\n    >>> echo.call("Goodbye, world", capture=True)\n    0\n    >>> echo.stdout()\n    [\'Hello, world\', \'Goodbye, world\']\n    >>> echo.stdout()\n    []\n\nPipe stdout to file with the ``file`` keyword argument\n\n.. code-block:: python\n\n    >>> import os\n    >>> import tempfile\n    >>>\n    >>> tmp = tempfile.NamedTemporaryFile(delete=False)\n    >>> echo.call("Hello, world", file=tmp.name)\n    0\n    >>> returncode = cat.call(tmp.name)\n    Hello, world\n    >>> returncode\n    0\n    >>> os.remove(tmp.name)\n\n    # redirect to /dev/null\n    >>> echo.call("Hello, world", file=os.devnull)\n    0\n\nFailing command will raise a ``subprocess.CalledProcessError``\n\n.. code-block:: python\n\n    >>> import contextlib\n    >>> from subprocess import CalledProcessError\n    >>>\n    >>> with contextlib.redirect_stderr(None):\n    ...     try:\n    ...         returncode = fails.call()\n    ...     except CalledProcessError as err:\n    ...         str(err)\n    "Command \'false\' returned non-zero exit status 1."\n    >>> returncode\n    0\n\nThis, however, will not\n\n.. code-block:: python\n\n    >>> with contextlib.redirect_stderr(None):\n    ...     fails.call(suppress=True)\n    1\n\nAll the keyword arguments above can be set as the default for the instantiated object\n\n.. code-block:: python\n\n    >>> echo = Subprocess("echo", capture=True)\n    >>> echo.call("Hello, world")\n    0\n    >>> echo.stdout()\n    [\'Hello, world\']\n\nWhich can then be overridden\n\n.. code-block:: python\n\n    >>> returncode = echo.call("Hello, world", capture=False)\n    Hello, world\n    >>> returncode\n    0\n    >>> echo.stdout()\n    []\n',
    'author': 'jshwi',
    'author_email': 'stephen@jshwisolutions.com',
    'maintainer': 'jshwi',
    'maintainer_email': 'stephen@jshwisolutions.com',
    'url': 'https://pypi.org/project/spall/',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
